### Main client game module

import sys
import pygame
from pygame.locals import *
from pgu import gui,text
import server_stub

dm_mode = False
server_name = ''
ready = False
pid = None

debug = 0

# Size of the client window
screenwidth = 800
screenheight = 600
screensize = (screenwidth, screenheight)

# Size of viewable map
viewsize = 600
panelsize = screenwidth - viewsize

pygame.init()

screen = pygame.display.set_mode(screensize)

lobby_background = pygame.image.load('images/lobby.bmp').convert()

def start(ignore):
    global ready, server_name, dm_mode, server_input, interface_group, http_handle, pid
    server_name = server_input.value

    error_box.value = "Checking hostname..."
    draw_lobby()
    error = server_stub.set_hostname(server_name)
    if error:
        error_box.value = error
        return
    error_box.value = "Connection Suceeded!"
    if interface_group.value == "dm":
        if server_stub.get_dm():
            dm_mode = True
            pid = -1
        else:
            error_box.value = "A Dungeon Master has already signed in for this session."
            interface_group.value = "knight"
            return
    else:
        pid = server_stub.get_pc(interface_group.value)
    ready = True

lobby_app = gui.App()

start_button = gui.Button("Start")
server_input = gui.Input(name="server_name",value="localhost",size=40)
error_box = gui.Label("")
interface_group = gui.Group(name="interface",value="knight")

start_button.connect(gui.CLICK,start,None)

lobby_table = gui.Table(x=50,y=150)
lobby_table.tr()
lobby_table.add(gui.Label("Server hostname: ",color=(0,0,0)))
lobby_table.add(server_input)
lobby_table.tr()
lobby_table.add(error_box)
lobby_table.tr()
lobby_table.add(gui.Radio(interface_group,value="knight"))
lobby_table.add(gui.Label(" play as Knight",align=-1))
lobby_table.tr()
lobby_table.add(gui.Radio(interface_group,value="mage"))
lobby_table.add(gui.Label(" play as Mage",align=-1))
lobby_table.tr()
lobby_table.add(gui.Radio(interface_group,value="elf"))
lobby_table.add(gui.Label(" play as Elf",align=-1))
lobby_table.tr()
lobby_table.add(gui.Radio(interface_group,value="dm"))
lobby_table.add(gui.Label(" play as Dungeon Master",align=-1))
lobby_table.tr()
lobby_table.add(start_button)

lobby_app.init(widget=lobby_table,screen=screen)

error_box.rect.w = 500
lobby_table.rect.w = screenwidth - lobby_table.rect.x
lobby_table.rect.h = screenheight - lobby_table.rect.y

def draw_lobby():
    screen.blit(lobby_background,(0,0))
    lobby_app.paint(screen)
    pygame.display.flip()

while not ready:
    draw_lobby()
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            sys.exit(0)
        elif event.type == KEYDOWN and event.key == K_RETURN:
            start(server_input)
        lobby_app.event(event)

# Clean up all the useless lobby UI objects
del lobby_background, lobby_table, lobby_app,
del interface_group, server_input, start_button

grid = server_stub.get_grid()
things, chatlines, positions, who_won = server_stub.get_state()

if dm_mode:
    import dm_interface
    interface = dm_interface.DMInterface()
else:
    import pc_interface
    interface = pc_interface.PlayerInterface(things[pid].position)
