### Client side

import pygame
import client

mapsize = 64
tilesize = 32

icons = {}

select_box = pygame.Surface((tilesize,tilesize))
select_box.set_colorkey((0,0,0))
pygame.draw.polygon(select_box,(255,0,0),[(0,0),
                                          (0,tilesize-1),
                                          (tilesize-1,tilesize-1),
                                          (tilesize-1,0)], 4)

class Icon(pygame.sprite.Sprite):

    def __init__(self,id,image):

        pygame.sprite.Sprite.__init__(self)

        self.id = id

        self.selected = False

        icon = pygame.image.load(image).convert()
        self.raw_image = pygame.transform.rotozoom(icon,0,float(tilesize)/icon.get_width())

        self.image = self.raw_image.copy()
        
        self.rect = pygame.Rect(0,0,self.image.get_width(),
                                self.image.get_height())

    def select(self):
        if self.selected: return
        self.selected = True
        self.image.blit(select_box,(0,0))

    def unselect(self):
        if not self.select: return
        self.selected = False
        self.image = self.raw_image.copy()

class Canvas:

    def __init__(self,position):

        self.rect = pygame.Rect(position[0],position[1],
                                client.viewsize,client.viewsize)
        
        self.surface = pygame.Surface((mapsize*tilesize,
                                       mapsize*tilesize))
        self.scratch = pygame.Surface((mapsize*tilesize,
                                       mapsize*tilesize))

        self.allsprites = pygame.sprite.RenderPlain()

        self.selected_rect = None

        for y in range(mapsize):
            for x in range(mapsize):
                id = client.grid[x][y].id
                self.scratch.blit(tiles[id],(tilesize*x,
                                                   tilesize*y))

        self.surface.blit(self.scratch,(0,0))
        self.update()

    def get_absolute_pos(self,x,y):
        return (x+self.rect.x)/tilesize,(y+self.rect.y)/tilesize
    
    def start_rect(self,x,y):
        x,y = self.get_absolute_pos(x,y)
        self.selected_rect = pygame.Rect(x,y,1,1)

    def select_rect(self,x,y):
        if x > self.rect.w - 2*tilesize:
            self.move(tilesize,0)
        if x < 2*tilesize:
            self.move(-tilesize,0)
        if y > self.rect.h - 2*tilesize:
            self.move(0,tilesize)
        if y < 2*tilesize:
            self.move(0,-tilesize)
        x,y = self.get_absolute_pos(x,y)
        self.selected_rect.w = x - self.selected_rect.x
        self.selected_rect.h = y - self.selected_rect.y

    def unselect_rect(self):
        self.selected_rect = None

    def get_thing_at_position(self,x,y):
        if x > self.rect.w: return None
        x,y = self.get_absolute_pos(x,y)
        try: return client.positions[(x,y)]
        except KeyError: return None

    def select_icon(self,thing):
        try: icons[thing.id].select()
        except KeyError: pass

    def unselect_icon(self,thing):
        try: icons[thing.id].unselect()
        except KeyError: pass

    def update(self):
        self.allsprites.clear(self.surface,self.scratch)
        for thing in client.things.values():
            icon = icons.get(thing.id)
            if not icon:
                icon = Icon(thing.id,'images/%s.bmp'%thing.type)
                icons[thing.id] = icon
            icon.rect.x = tilesize*thing.position[0]
            icon.rect.y = tilesize*thing.position[1]
            self.allsprites.add(icon)
        for icon in self.allsprites.sprites():
            if not client.things.has_key(icon.id): icon.kill()
        self.allsprites.update()
        self.allsprites.draw(self.surface)
                        
    def draw(self,screen):
        self.update()
        screen.blit(self.surface,(0,0),self.rect)
        if self.selected_rect:
            pygame.draw.rect(screen,(255,0,0),
                             (self.selected_rect.x*tilesize-self.rect.x,
                              self.selected_rect.y*tilesize-self.rect.y,
                              self.selected_rect.w*tilesize,
                              self.selected_rect.h*tilesize), 2)

    def move(self,x,y):
        self.rect.x = min(max(self.rect.x + x,0),
                          self.surface.get_width()-client.viewsize)
        self.rect.y = min(max(self.rect.y + y,0),
                          self.surface.get_height()-client.viewsize)

    def get_icon(self,x,y):
        x = x/tilesize
        y = y/tilesize
        for thing in client.things.values():
            if thing.position == (x,y):
                icon = icons[thing.id]
                icon.image = tiles['skull']
                return thing
        return None

import tiny_map

class sidePanel:
     
    color = (220,220,200)
     
    def __init__(self,width,height):
        self.surface = pygame.Surface((width,height))
        self.map = tiny_map.TinyMap(client.grid)

    def set_background(self, color):
        self.color = color
        self.surface.fill(color)
         
    def get_background(self):
        return self.color
         
    def erase(self, rect):
        self.surface.fill(self.color,rect)
     
    def replace_top(self, item):
        self.erase((0,0,200,50))
        x = (self.surface.get_width()-item.get_width())/2
        self.surface.blit(item,(x,0))
     
    def replace_middle(self, item):
        self.erase((0,50,200,50))
        x = (self.surface.get_width()-item.get_width())/2
        self.surface.blit(item,(x,50))

    def replace_middle2(self, item):
        self.erase((0,100,200,100))
        x = (self.surface.get_width()-item.get_width())/2
        self.surface.blit(item,(x,100))
         
    def replace_middle3(self, item):
        self.erase((0,200,200,250))
        x = (self.surface.get_width()-item.get_width())/2
        self.surface.blit(item,(x,200))
     
    def replace_bottom(self, item):
        self.erase((0,450,200,50))
        x = (self.surface.get_width()-item.get_width())/2
        self.surface.blit(item,(x,450))

    def draw(self, screen):
        screen.blit(self.surface,(client.viewsize,0))


tiles = {}

def load_tile(filename):
    if not filename: return pygame.Surface((tilesize,tilesize))
    image = pygame.image.load(filename).convert()
    return pygame.transform.rotozoom(image,0,float(tilesize)/image.get_width())

def load_tiles():
    import tile_list
    global tiles
    for number, filename in tile_list.tiles.items():
        tiles[number] = load_tile(filename)
    
