
import pygame
from pygame.locals import *
from pgu import gui,text
import sys

import menu
import client
import server_stub

def send_chat(textbox):
    # XXX Get rid of the cursor when not chatting.
    # XXX Chat is broken for the DM. We probably need to move the chat into
    # the interface class.
    server_stub.chat(client.pid,textbox.value)
    textbox.value = "press T to chat with other players"

app = gui.App()

# Setup the chat msg lines and chat input box

pid_chat_colors = {}
available_colors = [(0,0,0),
                    (125,0,0),
                    (0,125,0),
                    (0,0,125),
                    (125,125,0),
                    (0,125,125),
                    (125,0,125)]

table = gui.Table(x=0,y=client.screenheight-95)
chat_lines = [gui.Label(''),gui.Label(''),gui.Label('')]
chatInput = gui.Input(size=73,value="press T to chat with other players")
table.tr()
for line in chat_lines:
    table.add(line)
    table.tr()
table.add(chatInput)
app.init(widget=table,screen=client.screen)
table.rect.w = client.screenwidth - table.rect.x
table.rect.h = client.screenheight - table.rect.y

def update_chat():
    i = 0
    global pid_chat_colors, chat_lines
    for pid, line in client.chatlines:
        color = pid_chat_colors.get(pid)
        if not color:
            try:
                color = available_colors[0]
                available_colors.remove(color)
            except IndexError:
                color = (0,0,0)
            pid_chat_colors[pid] = color
        chat_lines[i].rect.x = 0
        chat_lines[i].rect.w = client.screenwidth - client.panelsize
        if not pid: type = ""
        elif pid == "-1": type = "Dungeon Master: "
        else: type = client.things[int(pid)].type + ": "
        line = type + line
        chat_lines[i].value = line
        chat_lines[i].style.color = color
        i += 1

class Interface:

    def __init__(self):

        self.update_time = 0
        self.battle_countdown = 0
        self.chatting = False
        self.clock = pygame.time.Clock()

    def draw_everything(self):
        app.paint(client.screen)
        if client.who_won:
            won_app = gui.App()
            table = gui.Table(x=200, y=200)
            table.tr()
            if client.who_won == "dm":
                table.add(gui.Label("Dungeon Master Wins!"))
            else:
                table.add(gui.Label("Player Wins!"))
            won_app.init(widget = table, screen = client.screen)
            table.rect.w = client.screenwidth - table.rect.x
            table.rect.h = client.screenheight - table.rect.y
            won_app.paint(client.screen)
        pygame.display.flip()


    def handle_event(self,event):
        if event.type == QUIT or (event.type == KEYDOWN and
                                    event.key == K_ESCAPE):
            sys.exit(0)
        if self.chatting and event.type == KEYDOWN:
            if event.key == K_RETURN:
                self.chatting = False
                send_chat(chatInput)
            elif not event.key in (K_UP,
                                   K_DOWN,
                                   K_LEFT,
                                   K_RIGHT):
                app.event(event)
            return -1
        if menu.active_menu:
            if event.type == MOUSEBUTTONDOWN and event.button == 1:
                action = menu.active_menu.handle_mouse_event(event.pos)
            else:
                action = menu.active_menu.handle_event(event)
            if action:
                return action
        elif event.type == KEYDOWN and event.key == K_t:
            self.chatting = True
            chatInput.value = ""
            return None
        return None

    def handle_keys_pressed(self,keys_down):
        return None

    def command_ok(self,command):
        return True

    def update_game_state(self,command):
        client.things, client.chatlines, client.positions, client.who_won = \
                       server_stub.get_state()

    def run(self):
    
        current_command = None

        while True:

            self.update_everything()
            self.draw_everything()

            # 1. Tick the clock
            tick_time = self.clock.tick(60)
            self.update_time += tick_time
            if self.battle_countdown > 0:
                self.battle_countdown -= tick_time
                if self.battle_countdown < 0:
                    self.battle_countdown = 0

            # 2. Read the user input from the keyboard, storing it in a
            # "last user input" buffer.
            new_command = None
            for event in pygame.event.get():
                new_command = self.handle_event(event)

            if not new_command:
                keys_down = pygame.key.get_pressed()
                new_command = self.handle_keys_pressed(keys_down)
            if new_command and new_command != -1:
                current_command = new_command

            # 3. Has x number of ticks passed since we last updated the game
            # state from the server?                
            if self.update_time > 200: # XXX At the moment, as fast as possible.
                self.update_time = 0

                if self.command_ok(current_command):
                    self.update_game_state(current_command)
                    current_command = None

            # 5. Calculate movement of all sprites to merge them toward the
            # current game state. 
            update_chat()

                

