### Client side module

import pygame
import client
import sound
from sound import sounds
import menu
import things
from pgu import gui,text

mapsize = 64
tilesize = 64

panelcolor = ((199,0,0))

icons = {}

player1 = None

class Icon(pygame.sprite.Sprite):
    """Generic class for any player, monster, or item that is rendered to
    the client."""

    def __init__(self,id,images,position):
        pygame.sprite.Sprite.__init__(self)

        self.id = id
        self.murder_me = isinstance(client.things[id],things.Creature)
        self.is_player = isinstance(client.things[id],things.Player)
        self.is_item = isinstance(client.things[id],things.Item)
        self.images = {}
        for name, image in images.items():
            try:
                self.images[name] = pygame.image.load(image).convert()
            except:
                self.images[name] = self.images['standing']
            #self.images[name].set_colorkey((255,255,255))
        self.image = self.images['standing']
        self.rect = pygame.Rect(position+(self.image.get_width(),
                                          self.image.get_height()))
        self.dead = False
        self.moving = False
        self.walk_frame = 1
        self.engaged = False
        self.damaged = None

    def murder(self):
        if not self.murder_me:
            if self.is_item and sound.on:
                sounds["wewon"].play()
            self.kill()
            return
        if self.dead: return
        self.dead = True
        self.engaged = False
        self.moving = False
        if sound.on:
            if self.is_player:
                sounds["dying"].play()
            else:
                sounds["eeeargh"].play()

    def move_to(self,x,y):
        self.rect.x = x
        self.rect.y = y

    def show_damage(self, amount=0, start=False):
        if start:
            self.damaged = [40,amount]
        else:
            if not self.damaged: return
            self.damaged[0] -= 2
        self.image = self.image.copy()
        app = gui.App()
        table = gui.Table(x=5,y=self.damaged[0])
        if self.damaged[1]: table.add(gui.Label("-"+str(self.damaged[1])))
        else: table.add(gui.Label("miss"))
        app.init(widget=table,screen=self.image)
        table.rect.w = self.image.get_width() - table.rect.x
        table.rect.h = self.image.get_height() - table.rect.y
        app.paint(self.image)
        if not self.damaged[0]: self.damaged = None
            
    def update(self):
        if self.moving:
            try:
                self.image = self.images['walk1']
            except KeyError:
                self.image = self.images['standing']
        elif self.engaged:
            if self.walk_frame == 10: self.walk_frame = 1
            else: self.walk_frame += 1
            if self.walk_frame <= 5:
                try:
                    self.image = self.images['fight']
                except KeyError:
                    self.image = self.images['standing']
            else:
                self.image = self.images['standing']
        else:
            if self.dead:
                self.image = self.images['dead']
                alpha = self.image.get_alpha()
                if not alpha: self.image.set_alpha(255)
                else:
                    alpha = max(alpha-10,0)
                    self.image.set_alpha(alpha)
                    if alpha == 0: self.kill()
                    else: self.image.set_alpha(alpha)
            else:
                self.image = self.images['standing']
        if self.damaged: self.show_damage()

class Canvas:
    def __init__(self,position):

        self.tilewidth = client.viewsize/tilesize+2
        self.offset = self.tilewidth / 2

        self.surface = pygame.Surface((self.tilewidth*tilesize,
                                         self.tilewidth*tilesize))
        self.scratch = pygame.Surface((self.tilewidth*tilesize,
                                         self.tilewidth*tilesize))

        self.best_view = (self.surface.get_width() - client.viewsize) / 2

        self.rect = pygame.Rect(self.best_view,
                                self.best_view,
                                client.viewsize,client.viewsize)
            
        self.x, self.y = position

        self.player1 = Icon(client.pid,
                            {'standing':
                             'images/%s.bmp'%client.things[client.pid].type,
                             'walk1':
                             'images/%sw1.bmp'%client.things[client.pid].type,
                             'fight':
                             'images/%s_fight.bmp'%client.things[client.pid].type,
                             'dead': 'images/skull.bmp'},
                            ((client.viewsize-tilesize)/2,
                             (client.viewsize-tilesize)/2))
        global player1
        player1 = self.player1
        
        self.allsprites = pygame.sprite.RenderPlain((self.player1,))

        self.rate = 8
        self.threshold = 0
        self.render_canvas()

    def render_canvas(self):
        for j in range(self.y - self.offset,
                       self.y - self.offset + self.tilewidth):
            for i in range(self.x - self.offset,
                           self.x - self.offset + self.tilewidth):
                if i < 0 or j < 0 or i >= mapsize or j >= mapsize:
                    id = 0
                else:
                    id = client.grid[i][j].id
                global tiles
                self.surface.blit(tiles[id],
                                  ((i-(self.x - self.offset))*tilesize,
                                   (j-(self.y - self.offset))*tilesize))
        self.scratch.blit(self.surface,(0,0))

    def move(self,x,y):
        delta_x = self.x - x
        delta_y = self.y - y
        self.x = x
        self.y = y
        self.rect.x = self.rect.x + delta_x * tilesize
        self.rect.y = self.rect.y + delta_y * tilesize
        for icon in self.allsprites.sprites():
            if icon.id == client.pid: continue
            icon.rect.x = icon.rect.x + delta_x * tilesize
            icon.rect.y = icon.rect.y + delta_y * tilesize
        self.render_canvas()

    def move_closer(self,x,y,best_x,best_y):
        if x == best_x and y == best_y: return x,y,False
        if abs(x-best_x) > tilesize or abs(y-best_y) > tilesize:
            return best_x,best_y,False
        if x < best_x - self.rate:
            x += self.rate
        elif x > best_x + self.rate:
            x -= self.rate
        else:
            x = best_x
        if y < best_y - self.rate:
            y += self.rate
        elif y > best_y + self.rate:
            y -= self.rate
        else:
            y = best_y
        return x, y, not (x==best_x and y==best_y)

    def update_player(self):
        self.rect.x,self.rect.y,self.player1.moving = \
                                    self.move_closer(self.rect.x,self.rect.y,
                                                     self.best_view,
                                                     self.best_view)
        if client.things[client.pid].isDead(): self.player1.murder()
        else:
            self.player1.dead = False
            self.allsprites.add(self.player1)
        if client.things[client.pid].engaged(): self.player1.engaged = True
        else: self.player1.engaged = False
#        return abs(self.rect.x - self.best_view) > self.threshold or \
#               abs(self.rect.y - self.best_view) > self.threshold

    def move_sprites(self):
        self.allsprites.clear(self.surface,self.scratch)
        self.player1.rect.x = self.rect.x + (client.viewsize - tilesize) / 2
        self.player1.rect.y = self.rect.y + (client.viewsize - tilesize) / 2
        for id, thing in client.things.items():
            if id == client.pid: continue

            icon = icons.get(id)

            x, y = thing.position
            if (abs(x - self.x) < self.tilewidth and 
                abs(y - self.y) < self.tilewidth):
                target_x = (x - (self.x - self.offset)) * tilesize
                target_y = (y - (self.y - self.offset)) * tilesize
                if icon is None:
                    icon = Icon(id,{'standing':'images/%s.bmp'%thing.type,
                                    'walk1':
                                    'images/%sw1.bmp'%thing.type,
                                    'fight':
                                    'images/%s_fight.bmp'%thing.type,
                                    'dead':'images/skull.bmp'},
                                (target_x,target_y))
                    icons[id] = icon

                if isinstance( thing, things.Crystal ):
                    if thing.owned():
                        bmp = thing.get_image()
                        icon.image = tiles[ bmp ] 
                        icon.images['standing'] = tiles[ bmp ] 
                        icon.images['walk1'] = tiles[ bmp ] 
            
                new_x, new_y, icon.moving = self.move_closer(icon.rect.x,
                                                             icon.rect.y,
                                                             target_x,
                                                             target_y)
                if isinstance(thing,things.Creature):
                    if thing.engaged(): icon.engaged = True
                    else: icon.engaged = False
                icon.move_to(new_x,new_y)
                icon.add(self.allsprites)
            else:
                if icon and icon.alive():
                    icon.remove(self.allsprites)
        for icon in self.allsprites.sprites():
            if not client.things.has_key(icon.id): icon.murder()
        self.allsprites.update()
        self.allsprites.draw(self.surface)

    def draw(self,screen):
        self.move_sprites()
        screen.blit(self.surface,(0,0),self.rect)

    def moveto(self,x,y):
        if self.x != x or self.y != y:
            self.move(x,y)
        self.update_player()

class CrystalBar:
    def __init__(self,pid):
        self.surface = pygame.Surface((80,20))  # size of the thing wide/high
        self.pid = pid
        self.update()

    def update(self):
        self.draw_crystals_owned()

    def draw_crystals_owned( self ):
        owned_panel = self.surface 
        me = client.things[self.pid]

        bar = 0
        if ( me.id == client.pid ): # my window?
            owned_panel.fill((255,255,255),(0,0,80,20)) # draw the background of the bar
            for x in range( len( me.crystals ) ):
                crystalid = me.crystals[x]
                crystal = client.things[ long( crystalid ) ]
                hex = crystal.get_color_hex()
                #print "color of crystal is ", hex
                owned_panel.fill(hex,(bar,0,20,20)) # draw the crystal and it's color
                bar = bar + 20
                 
            
class LifeBar:

    def __init__(self,pid):
        self.surface = pygame.Surface((170,20))
        self.pid = pid
        self.update()

    def update(self):
        self.draw_life()

    def draw_life(self):
        # XXX Creating a new surface with every frame may be slow.
        life = self.surface
        creature = client.things[self.pid]
        if creature.id == client.pid:
            if creature.isDead():
                icon = pygame.transform.scale((player1.images["dead"]),(20,20))
                bar = 0
            else:
                bar = (creature.hp*150)/creature.max_hp
                if creature.engaged():
                    icon = pygame.transform.scale((player1.images["fight"]),(20,20))
                else:    
                    icon = pygame.transform.scale((player1.images["standing"]),(20,20))
        else:
            if creature.isDead():
                icon = pygame.transform.scale((icons[creature.id].images["dead"]),(20,20))
                bar = 0
            else:
                bar = (creature.hp*150)/creature.max_hp
                if creature.engaged(): # XXX only checks engaged with this PC
                    icon = pygame.transform.scale((icons[creature.id].images["fight"]),(20,20))
                else:
                    icon = pygame.transform.scale((icons[creature.id].images["standing"]),(20,20))    
        life.blit(icon,(0,0,20,20))
        life.fill((255,0,0),(20,0,170,20))
        life.fill((0,255,0),(20,0,bar,20))

class MagicBar:

    def __init__(self,pid):
        self.surface = pygame.Surface((170,20))
        self.pid = pid
        self.icon = pygame.transform.scale(pygame.image.load('images/wand.bmp').convert(), (20, 20))
        self.update()

    def update(self):
        self.draw_magic()

    def draw_magic(self):
        # XXX Creating a new surface with every frame may be slow.
        magic = self.surface
        creature = client.things[self.pid]
        bar = (creature.mp*150)/creature.max_mp
        magic.blit(self.icon,(0,0,20,20))
        magic.fill((255,0,0),(20,0,170,20))
        magic.fill((0,255,0),(20,0,bar,20))

import tiny_map

class subPanel(pygame.sprite.Sprite):

    def __init__(self,image,rect):
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = pygame.Rect(rect)

class sidePanel:
     
    def __init__(self,width,height):
        self.surface = pygame.Surface((width,height))
        self.color = panelcolor
     
        self.surface.fill(self.color)

        self.map = tiny_map.TinyMap(client.grid)

        self.lifebar = LifeBar(client.pid)
        self.magicbar = MagicBar(client.pid)
        self.crystalbar = CrystalBar(client.pid)

        self.regions = {'top' : 0,
                        'middle1': 50,
                        'middle1andahalf': 75,
                        'middle2': 100,
                        'middle3': 200,
                        'bottom' : 450,
                        }

        self.sprites = pygame.sprite.OrderedUpdates()

        self.sprites.add(subPanel(self.surface,(client.viewsize,0,
                                                width,height)))

        logo = pygame.image.load('images/pq.bmp').convert()
        logo.set_colorkey((199,0,0))
        self.add_image(logo,'top')
        self.add_image(self.map.surface,'bottom')
        self.add_image(self.lifebar.surface,'middle1')
        self.add_image(self.magicbar.surface,'middle1andahalf')
        self.add_image(self.crystalbar.surface,'middle2')

    def center_in_panel(self,image):
        return client.viewsize + (client.panelsize - image.get_width()) / 2

    def add_image(self,image,region):
        self.sprites.add(subPanel(image,(self.center_in_panel(image),
                                         self.regions[region],
                                         image.get_width(),
                                         image.get_height())))

    def add_panel(self,subpanel,region,center=True):
        if center:
            subpanel.rect.x = self.center_in_panel(subpanel.image)
        else:
            subpanel.rect.x += client.viewsize
        subpanel.rect.y = self.regions[region]
        self.sprites.add(subpanel)

    def remove_panel(self,subpanel):
        subpanel.rect.x -= client.viewsize
        self.sprites.remove(subpanel)

    def draw(self, screen):
        self.sprites.draw(screen)

tiles = {}

def load_tile(filename):
    if not filename: return pygame.Surface((tilesize,tilesize))
    return pygame.image.load(filename).convert()

def load_tiles():
    import tile_list
    global tiles
    for number, filename in tile_list.tiles.items():
        tiles[number] = load_tile(filename)
    
