
### thing to generate a random map/dungeon
### this is a server side module

import pygame

import random
import string
import time

import things

# innercorners are:  upper left 55  upper right 56   lower left 59   lower right 60


# beginnings of a dungeon diagram drawing function
# still awfully buggy

#              column y 
#   row  _________________________              
#    x   | | | | | | | | | | | | | 
#        +~+~+~+~+~+~+~+~+~+~+~+~+
#        | | | | | | | | | | | |@|  (row,col) = ( 2, 12 ) = ( x, y ) 
#        +~+~+~+~+~+~+~+~+~+~+~+~+

# please leave this debugging stuff here please -JL

#import pdb
#import foo
#from foo import Tile
#from foo import Foo 
#things = Foo()


class LinePair:
	def __init__(self, line, val ):
		self.line = line
		self.val = val

class Map:

	# parameters are:  rows, columns,
	# sparseness:  floating point number meaning at what point should it stop drawing
	#              e.g. 0.20 - stop drawing when there's 20% of the unvisited map left to explore
	# turn_frequency:  randomness factor which lets you control how often it decides to turn
	#              this number should be an integer, e.g. 4, 7, 10, etc.
	#              the larger the number, the straighter the maze paths will be
	# rooms        max number of rooms ( randomly chosen )
	# minrooms     minimum number of rooms
	# max_room_width  

	def __init__( self, rows, cols, sparseness, turn_frequency, rooms, minrooms, max_room_width ):
		self.m_nrows = rows
		self.m_ncols = cols
		self.m_map = []
		self.m_fsparseness = sparseness
		self.m_nturn_frequency = turn_frequency
		for col in range( cols ):
			self.m_map.append( [] )
			for row in range( rows ):
				self.m_map[col].append( None )

		corridor_width = 3

		self.m_nrooms = random.randint(minrooms,rooms)
		self.m_nmax_room_width = max_room_width
#		if ( self.m_nmax_room_width < corridor_width + 4 ):
#			self.m_nmax_room_width = corridor_width + 4

	def is_free_square( self, row, col ):
		return (self.is_within_map( row, col ) and self.m_map[col][row] is None)

	def is_unvisited( self, row, col ):
		return self.is_within_map( row, col ) and self.is_free_square( row, col )		

	def is_unvisited_point( self, point ):
		return self.is_unvisited( point[0], point[1] )

	def is_point_within_map( self, point ):
		return self.is_within_map( point[0], point[1] )

	def is_outer_wall( self, row, col ):
    	# don't call without checking for is_within_map first
		return self.m_map[col][row] and self.m_map[col][row].id == 51

	def is_unvisited_direction( self, row, col, direction ):
		# test the direction to see if it is occupied or not
		myrow, mycol = self.adjust_for_direction( row, col, direction )
		return self.is_unvisited( myrow, mycol )

	def is_within_map( self, row, col ):
		if ( row >= self.m_nrows or row < 0 ): 		# hit the edge of the map
			return False
		if ( col >= self.m_ncols or col < 0 ):		# hit the edge of the map
			return False
		if self.is_outer_wall( row, col ):
			return False 
		return True

	def is_within( self, row, col, rooms ):
		for x in rooms:
			if ( row > x[0] and row < x[2] and col > x[1] and col < x[3] ): 
				return True
		return False 
		
	def get_other_path_points(self, rooms):
		# first create a list of the available points
		things = []

		for row in range(self.m_nrows):
			for col in range(self.m_ncols):
				if ( self.m_map[col][row] and self.m_map[col][row].id == 1 ): # walkable path 
					if ( not self.is_within( row, col, rooms ) ):
						things.append( [row, col] )
		return things

	def get_path_points(self):
		# first create a list of the available points
		things = []

		for row in range(self.m_nrows):
			for col in range(self.m_ncols):
				if ( self.m_map[col][row] and self.m_map[col][row].id == 1 ): # walkable path 
					things.append( [row, col] )
		return things

	def get_unvisited_points(self):
		# first create a list of the available points
		things = []

		for row in range(self.m_nrows):
			for col in range(self.m_ncols):
				if ( self.m_map[col][row] is None ):
					things.append( [row, col] )	
		return things

	def get_all_points(self):
		things = []
		for row in range(self.m_nrows):
			for col in range(self.m_ncols):
				things.append( [row, col] )	
		return things

	def count_points(self):
		return self.m_nrows * self.m_ncols

	def mark2( self, row, col, tile ):	
		point = [row, col]
		if ( self.is_point_within_map( point ) ):
			self.m_map[point[1]][point[0]] = tile

	def mark( self, point, tile ):
		if ( self.is_unvisited_point( point ) ):
			self.m_map[point[1]][point[0]] = tile

	def is_valid_turning_point( self, row, col, direction, path ):
		if ( not self.is_within_map( row, col ) ):
			return False

		if ( self.is_free_square( row, col ) == True ):		# wide open 
			return True

		# if we are turning, if it is a wall, either horizontal or vertical and the space I will walk into is 
		# unvisited, then it's okay

		#   east->north
		#   west->north
		#                        0
		#                      ---
        #                        . 
        #                      ---

		#   east->south
		#   west->south
		#                      ---
        #                        . 
        #                      ---
		#                        0

		#   north->east
		#   south->east
		#                      |.|0
        #                      | |

		#   north->west
		#   south->west
		#                     0|.|
        #                      | |

		if ( path == "horizontal" ):
			if ( direction == "north" ): # should always be 90 degrees when changing directions
				if ( self.is_free_square( row - 1, col ) == True ):
					return True
				else:
					return False
			else:
				if ( self.is_free_square( row + 1, col ) == True ):
					return True
				else:
					return False
		else:
			if ( direction == "east" ):
				if ( self.is_free_square( row, col+1 ) == True ):
					return True
				else:
					return False
			else:
				if ( self.is_free_square( row, col-1 ) == True ):
					return True
				else:
					return False			
		return False 

    # this needs to be adjusted for wider corridors 
	def adjust_for_direction( self, yourrow, yourcol, direction ):
		row = yourrow
		col = yourcol
		if ( direction == "north" ):
			row = row - 1
		elif ( direction == "south" ):
			row = row + 1
		elif ( direction == "east" ):
			col = col + 1
		elif ( direction == "west" ):
			col = col - 1

		return row, col 


	def blast_through_a_wall(self, row, col, direction, width ):
		# this is like pick a direction, only we want to go in a direction that is not an existing path
		# which means we'll have to break a wall down
		# look at all squares, where there is a path, we can't go there, what's left is what we can choose from

		choices = ["north", "south", "east", "west"] 

		# what we are trying to do here, is we don't want to traverse a path
		# so we want to take the narrowest option and look there
		leftw, rightw, left, right = self.get_path_width( row, col, "horizontal" )
		upw, downw, up, down = self.get_path_width( row, col, "vertical" )

		# take the narrowest direction and discard the rest

		horiz = leftw+ rightw
		vert = upw + downw

		otherchoices = choices[:]
		if ( horiz < vert ):
			choices.remove( "north", )
			choices.remove( "south" )
			otherchoices.remove( "east" )
			otherchoices.remove( "west" )
		else:
			choices.remove( "east" )
			choices.remove( "west" )
			otherchoices.remove( "north" )
			otherchoices.remove( "south" )
	
		# of the remaining choices, which one is better?  it's good if on the other side there is unvisited space
		# if neither side is unvisited.....  I guess we have a dead end?
		direction = random.choice( choices )
		ShouldBlast, row, col = self.is_good_for_blasting( row, col, direction, width ) # a good direction
		if ( ShouldBlast ):
			return self.blast( direction, row, col, width )

		choices.remove( direction )
		if ( choices == [] ):
			return False, -1, -1, "none" 

		direction = random.choice( choices )
		ShouldBlast, row, col = self.is_good_for_blasting( row, col, direction, width ) # a good direction
		if ( ShouldBlast ):
			return self.blast( direction, row, col, width )

		# if we get here,, what the heck, try the other directions...
		direction = random.choice( otherchoices )
		ShouldBlast, row, col = self.is_good_for_blasting( row, col, direction, width ) # a good direction
		if ( ShouldBlast ):
			return self.blast( direction, row, col, width )

		otherchoices.remove( direction )

		direction = random.choice( otherchoices )
		ShouldBlast, row, col = self.is_good_for_blasting( row, col, direction, width ) # a good direction
		if ( ShouldBlast ):
			return self.blast( direction, row, col, width )
		
		return False, -1, -1, "none" 

	def adjust_negative ( self, num ):
		if ( num < 0 ): 
			num = 0
		return num

	def adjust_limit( self, num, orient ):
		num = self.adjust_negative( num )
		if ( orient == "north" or orient == "south" ):
			orient = "vertical"
		if ( orient == "east" or orient == "west" ):
			orient = "horizontal"
		
		if ( orient == "horizontal" and num > self.m_ncols ): 
			num = self.m_ncols - 1
		if ( orient == "vertical" and num > self.m_nrows ): 
			num = self.m_nrows - 1
		return num

	def get_path_width( self, row, col, orient ):
		if ( orient == "horizontal" ):
			for x in range( self.m_ncols ):
				if ( not self.is_path( row, col - x ) ):
					break
			left = col - x
			for x in range( self.m_ncols ):
				if ( not self.is_path( row, col + x ) ):
					break
			right = col + x
			leftw = col - left
			rightw = right - col

			left =  self.adjust_limit( left, orient )
			right = self.adjust_limit( right, orient ) 
			leftw = self.adjust_negative( leftw ) 
			rightw = self.adjust_negative( rightw )
			
			return leftw, rightw, left, right

		if ( orient == "vertical" ):
			for x in range( self.m_ncols ):
				if ( not self.is_path( row - x, col ) ):
					break
			up = row - x
			for x in range( self.m_ncols ):
				if ( not self.is_path( row + x, col ) ):
					break
			down = row + x
			upw = row - up
			downw = down - row

			up = self.adjust_limit( up, orient )
			down = self.adjust_limit( down, orient )
			upw = self.adjust_negative( upw )
			down = self.adjust_negative( downw )

			return upw, downw, up, down 

	def get_close_to_wall( self, row, col, direction ):
		if ( row == -1 or col == -1 ):
			return row, col 
		myrow = row 
		mycol = col 
		if ( direction == "north" ): 
			for x in range(self.m_nrows):
				if ( not self.is_path( row - x, mycol ) ):
					myrow = row - x + 1
					break	
		if ( direction == "south" ): 
			for x in range(self.m_nrows):
				if ( not self.is_path( row + x, mycol ) ):
					myrow = row + x - 1
					break	
		if ( direction == "east" ): 
			for x in range(self.m_ncols):
				if ( not self.is_path( myrow, col + x ) ):
					mycol = col + x	 - 1 
					break	
		if ( direction == "west" ): 
			for x in range(self.m_ncols):
				if ( not self.is_path( myrow, col - x ) ):
					mycol = col - x	 + 1
					break	
		if ( myrow < 0 ): myrow = 0
		if ( mycol < 0 ): mycol = 0
		return myrow, mycol

	def center_blast( self, yourrow, yourcol, direction, width ):
		# re-center ourselves within the path - allow for 1/2 width of path on either side of blast direction
		row = yourrow
		col = yourcol
		myrow = yourrow
		mycol = yourcol
		
		half = width / 2  + 1
		# center across blast 
		# we only want to center across the blast if we are in a narrow passage, if the passage is wider than
		# the width of the corridor, don't bother
		
		if ( direction == "north" or direction == "south" ):
			leftw, rightw, left, right = self.get_path_width( row, col, "horizontal" )
			if ( leftw < half ):
				mycol = left + half
			if ( rightw < half ):
				mycol = right - half
		else:	
			upw, downw, up, down = self.get_path_width( row, col, "vertical" )
			if ( upw < half ):
				myrow = upw + half
			if ( downw < half ):
				myrow = downw - half 

		return self.get_close_to_wall( myrow, mycol, direction )	

	def blast( self, direction, yourrow, yourcol, width ):
		# power through this tunnel
		myrow, mycol = self.center_blast( yourrow, yourcol, direction, width )
		if ( myrow == -1 or mycol == -1 ): # error
			return False, -1, -1, "none"

		row = myrow 
		col = mycol 
		if ( direction == "north" ):
			for x in range(width):
				col = mycol
				col = col + x - 1	
				self.mark2( row, col, things.Tile(1) )
				self.mark2( row-1, col, things.Tile(1) )
				self.mark2(row-2, col, things.Tile(1) )
				self.mark2(row-3, col, things.Tile(1) )
			row = myrow
			col = mycol
			self.mark_walls(row-3, col, direction, width)
			self.mark_walls(row-2, col, direction, width)
			return True, row-4, col, direction
		if ( direction == "south" ):
			for x in range(width):
				col = mycol
				col = col + x - 1	
				self.mark2(row, col,  things.Tile(1) )
				self.mark2(row+1, col,  things.Tile(1) )
				self.mark2(row+2, col,  things.Tile(1) )
				self.mark2(row+3, col,  things.Tile(1) )
			row = myrow
			col = mycol
			self.mark_walls(row+3, col, direction, width)
			self.mark_walls(row+2, col, direction, width)
			return True, row+4, col, direction
		if ( direction == "east" ):
			for x in range(width):
				row = myrow
				row = row + x - 1	
				self.mark2(row, col,  things.Tile(1) )
				self.mark2(row, col+1,  things.Tile(1) )
				self.mark2(row, col+2,  things.Tile(1) )
				self.mark2(row, col+3,  things.Tile(1) )
			row = myrow
			col = mycol
			self.mark_walls(row, col+3, direction, width)
			self.mark_walls(row, col+2, direction, width)
			return True, row, col+4, direction
		if ( direction == "west" ):
			for x in range(width):
				row = myrow
				row = row + x - 1	
				self.mark2(row, col,  things.Tile(1) )
				self.mark2(row, col-1,  things.Tile(1) )
				self.mark2(row, col-2,  things.Tile(1) )
				self.mark2(row, col-3,  things.Tile(1) )
			row = myrow
			col = mycol
			self.mark_walls(row, col-3, direction, width)
			self.mark_walls(row, col-2, direction, width)
			return True, row, col-4, direction

		return False, -1, -1, "none"
		

	# if it is good for blasting, that means we want to find unvisited space on the other side
	def is_good_for_blasting( self, yourrow, yourcol, direction, width ):
		row, col = self.get_close_to_wall( yourrow, yourcol, direction )	
		if ( direction == "north" ): # look up 3 spaces, is it free?
			if ( self.is_unvisited( row - 3, col ) ):
				return True, row, col	
		if ( direction == "south" ): # look down 3 spaces, is it free?
			if ( self.is_unvisited( row + 3, col ) ):
				return True, row, col	
		if ( direction == "east" ): # look right 3 spaces, is it free?
			if ( self.is_unvisited( row, col + 3 ) ):
				return True, row, col	
		if ( direction == "west" ): # look right 3 spaces, is it free?
			if ( self.is_unvisited( row, col - 3 ) ):
				return True, row, col

		return False , -1, -1
									
	# awfully similar to is_good_for_blasting
	def make_tunnel( self, yourrow, yourcol, direction, width ):
		row, col = self.get_close_to_wall( yourrow, yourcol, direction )	
		if ( direction == "north" ): # look up 3 spaces, is it free?
			if ( self.is_path( row - 3, col ) ):
				return True, row, col	
		if ( direction == "south" ): # look down 3 spaces, is it free?
			if ( self.is_path( row + 3, col ) ):
				return True, row, col	
		if ( direction == "east" ): # look right 3 spaces, is it free?
			if ( self.is_path( row, col + 3 ) ):
				return True, row, col
		if ( direction == "west" ): # look right 3 spaces, is it free?
			if ( self.is_path( row, col - 3 ) ):
				return True, row, col

		return False, -1, -1

	def pick_a_direction(self, yourrow, yourcol, yourdir, width ):

		# to keep from ambling so crazily, allow the program to continue in a particular direction
		# for a while before changing it's mind, use a random chance of 0-some range before we change directions 
		keepgoing = random.randrange( 0, self.m_nturn_frequency, 1 )
		if ( yourdir != "none" and keepgoing > 0 ):
			if ( self.is_unvisited_direction( yourrow, yourcol, yourdir ) == True ):
				row, col = self.adjust_for_direction( yourrow, yourcol, yourdir )
				return True, row, col, yourdir 
		# if we are going to choose a new direction, like right now, we have to build a corner wall when we turn
		# whatever corner it is

		FoundAValidDirection, myrow, mycol, mydir = self.pick_a_random_direction( yourrow, yourcol, yourdir ) 
	
		# prevent dead-end-iness, blast through this wall, if possible, then return false as before
		if ( not FoundAValidDirection and yourdir != "none" ):
			self.make_tunnel( yourrow, yourcol, yourdir, width )

		return FoundAValidDirection, myrow, mycol, mydir
			

	def draw_corner_corridor( self, corridor_width, yourrow, yourcol, direction, destination ):
		# draw a block of (coridor_width) ^ 2 to the immediate north of my current position 
		point = [-1,-1]

		row = yourrow
		col = yourcol

		# draw the corridor path 
		for x in range(corridor_width ):
			for y in range( corridor_width ):
				if ( direction == "north" ):
					point = [row-x, col - 1 + y ]
				elif ( direction == "south" ):
					point = [row+x, col - 1 + y]
				elif ( direction == "east" ):
					point = [row - 1 +x, col+y]
				else: # direction == west 
					point = [row -1 + x, col-y]
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(1) )   


		width = corridor_width
		half = ( corridor_width / 2 )#  + (corridor_width % 2 )

		# build the opposite wall (facing your direction)
		for x in range( corridor_width ):
			row = yourrow
			col = yourcol
			if ( direction == "north" ):
				row = row - width 
				col = col - half + x
				point = [row, col]
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(57) ) 
				point = [row-1, col] # filler
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(64) ) 

			if ( direction == "south" ):
				row = row + width  
				col = col - half + x
				point = [row, col]
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(61) ) 
				point = [row+1, col] # filler
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(65) ) 

			if ( direction == "east" ):
				row = row - half + x
				col = col + width  
				point = [row, col]
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(62) ) 
				point = [row, col+1] # filler
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(65) ) 

			if ( direction == "west" ):
				row = row - half  + x
				col = col - width 
				point = [row, col]
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(63) ) 
				point = [row, col-1] # filler
				if ( self.is_point_within_map( point ) ):
					self.mark( point, things.Tile(65) ) 

		# build the side wall (opposite your destinatino)
		for x in range( corridor_width ):
			row = yourrow
			col = yourcol
			if ( destination == "north" ):
				row = row + half + 1
				row2 = row + 1 
				col2 = col
				tile = things.Tile( 61 )
			if ( destination == "south" ):
				row = row - half - 1
				row2 = row - 1 
				col2 = col
				tile = things.Tile( 57 )
			if ( destination == "east" ):
			 	col = col - half - 1	
				col2 = col - 1 
				row2 = row 
				tile = things.Tile( 62 )
			if ( destination == "west" ):
			 	col = col + half + 1
				col2 = col + 1 
				row2 = row 
				tile = things.Tile( 63 )

			if ( direction == "north" ):
				row = row - half - 1
			if ( direction == "south" ):
			  	row = row	
			if ( direction == "east" ):
			  	col = col	
			if ( direction == "west" ):
			  	col = col - width + 1	
				
			if ( destination == "north" ):
			 	col = col + x	
			if ( destination== "south" ):
			 	col = col + x	
			if ( destination== "east" ):
				row = row + x	
			if ( destination== "west" ):
				row = row + x	
			
			point = [row, col]
			if ( self.is_point_within_map( point ) ):
				self.mark( point, tile )   
			point = [row2, col2]
			if ( self.is_point_within_map( point ) ):
				self.mark( point, things.Tile(65) )   

		# draw the corner
		row = yourrow
		col = yourcol
		if ( destination == "north" ):
			row = row + half + 1
		if ( destination == "south" ):
			row = row - half - 1
		if ( destination == "east" ):
		 	col = col - half - 1	
		if ( destination == "west" ):
		 	col = col + half + 1

		if ( direction == "north" ):
			row = row - width 
		if ( direction == "south" ):
		  	row = row + width	 
		if ( direction == "east" ):
		  	col = col	+ width
		if ( direction == "west" ):
			col = col - width
			
		point = [row, col]
		if ( self.is_point_within_map( point ) ):
			self.mark( point, things.Tile(50) )   
		

		# adjust the new starting point
		row = yourrow
		col = yourcol
		if ( destination == "north" ):
			row = row - half - 1
		if ( destination == "south" ):
			row = row + half + 1
		if ( destination == "east" ):
		 	col = col + half + 1	
		if ( destination == "west" ):
		 	col = col - half - 1

		if ( direction == "north" ):
			row = row - half 
		if ( direction == "south" ):
			row = row + half
		if ( direction == "east" ):
		 	col = col + half
		if ( direction == "west" ):
		 	col = col - half
		

		return row, col


	def draw_corner( self, corridor_width, oldrow, oldcol, row, col, olddir, newdir ):
		# where is the corner?
		#  north->west?       -+
		#                      |
		#
		#  north->east?       +-
		#                     |
		#
		#  south->west?       |
		#                    -+
		#
		#  south->east?		  |
		#                     +-
		#
		#                     |
		#  east->north       -+ 
		#
		#  east->south       -+
		#                     |
		#
		#  west->north       |
		#                    +-
		#
		#  west->south       +-
		#                    |
		#                    

		#self.m_map[col][row] = things.Tile(1)             # new path

		row, col = self.draw_corner_corridor( corridor_width, oldrow, oldcol, olddir, newdir )
		# don't do this because we want an error value so we can start a new path
	#	row = self.adjust_limit( row, newdir )
	#	col = self.adjust_limit( row, newdir )
		return row, col
		
	
	def pick_a_random_direction( self, yourrow, yourcol, yourdir ):
		FoundAValidDirection = False

		nCounter = 4 
		mycol = -1
		myrow = -1
		mydir = yourdir 
		path = ""

		choices = ["north", "south", "east", "west" ]
		if ( yourdir == "south" or yourdir == "north" ): # go at some angle, not forward or backwards 
			choices.remove( "south" ) 
			choices.remove( "north" ) 
			path = "vertical"
		elif ( yourdir == "west" or yourdir == "east" ):
			choices.remove( "west" )
			choices.remove( "east" )
			path = "horizontal"
		while ( not FoundAValidDirection and nCounter > 0 ):	

			if ( choices == [] ):
				break

			direction = random.choice(choices)
			col = yourcol
			row = yourrow
			nCounter = nCounter - 1
			row, col = self.adjust_for_direction( row, col, direction )

			# don't try that direction again...
			choices.remove( direction ) #= choices - dir 
			
			#if I am turning a corner, the corridor walls I just built should be valid...

			if ( not self.is_valid_turning_point( row, col, direction, path ) ):
				continue

			# made it here without a continue, that's good

			FoundAValidDirection = True
			mycol = col 
			myrow = row 
			mydir = direction 
			break

		return FoundAValidDirection, myrow, mycol, mydir

	def mark_walls( self, row, col, direction, width ):
		mycol = col 
		myrow = row
		mydir = direction 
		# mark the walls - use dir to tell us whether to make horizontal or vertical walls
		if ( direction == "north" or direction == "south" ):
			mydir = "vertical"
		else:
			mydir = "horizontal"

		total_width = width + 4 # two fillers and two walls = 4
		halfway = total_width / 2 

		if ( mydir == "vertical" ):              # make walls like this |  | 
			filler1 = col - halfway
			wall1 = col - halfway + 1
			if ( self.is_unvisited( myrow, wall1 ) == True and not self.is_wall_point( myrow, wall1 ) ): 
				self.mark2( myrow, wall1, things.Tile(63) )
			if ( self.is_unvisited( myrow, filler1 ) == True and not self.is_wall_point( myrow, filler1 ) ):
				self.mark2( myrow, filler1, things.Tile(65) )

			for x in range( width ):
				add = wall1 + x + 1			
				if ( self.is_unvisited( myrow, add ) == True and not self.is_wall_point( myrow, add ) ): 
					self.mark2( myrow, add, things.Tile(1) )
			wall2 = add + 1 
			filler2 = wall2 + 1 
			if ( self.is_unvisited( myrow, wall2 ) == True and not self.is_wall_point( myrow, wall2 ) ):
				self.mark2( myrow, wall2, things.Tile(62) )
			if ( self.is_unvisited( myrow, filler2 ) and not self.is_wall_point( myrow, filler2 ) ):
				self.mark2( myrow, filler2, things.Tile(65) )

											#                       __
											# make walls like this     
											#                       __
		if ( mydir == "horizontal" ):        
			filler1 = row - halfway
			wall1 = row - halfway + 1
			if ( self.is_unvisited( wall1, mycol ) and not self.is_wall_point( wall1, mycol ) ):
				self.mark2( wall1, mycol, things.Tile(61) )
			if ( self.is_unvisited( filler1, mycol ) and not self.is_wall_point( filler1, mycol ) ):
				self.mark2( filler1, mycol, things.Tile(65) )

			for x in range( width ):
				add = wall1 + x	+ 1
				if ( self.is_unvisited( add, mycol ) == True and not self.is_wall_point( add, mycol ) ): 
					self.mark2( add, mycol, things.Tile(1) )
			wall2 = add + 1 
			filler2 = wall2 + 1 
			if ( self.is_unvisited( wall2, mycol ) and not self.is_wall_point( wall2, mycol ) ):
				self.mark2( wall2, mycol, things.Tile(57) )
			if ( self.is_unvisited( filler2, mycol ) and not self.is_wall_point( filler2, mycol ) ):
				self.mark2( filler2, mycol,  things.Tile(65) )
			
			
	def pick_a_random_path_point(self):
		things = self.get_path_points() 
		if ( things != [] ):
			point = random.choice( things )
			return point
		else:
			return [-1,-1]

	def pick_a_random_unvisited_point(self):
		things = self.get_unvisited_points() 
		if ( things != [] ):
			point = random.choice( things )
			return point
		else:
			return [-1,-1]

	# instead of trying for total completion - go for a percentage
	def map_completed( self ):
		things = self.get_unvisited_points()
		lenall = self.count_points()
		lenleft = len(things)
		left = float(lenleft) / float(lenall)
		return things == [] or left < self.m_fsparseness 


	# is this point a wall or edge of map or some other obstacle like that?
	def is_wall_point( self, row, col ):
		point = [row,col]
		return self.is_wall( point )

	def is_unbreachable( self, row, col ):
		return self.is_within_map( row, col ) and self.is_outer_wall( row, col ) 

	def is_path( self, row, col ):
		bool1 = self.is_within_map( row, col ) 
		if ( not bool1 ):
			return False
		bool2 = self.m_map[col][row] is None
		bool3 = False	
		if ( not bool2 ):
			bool3 = self.m_map[col][row].id == 1

		return bool1 and bool3


	def is_filler( self, point ):
		return (self.is_point_within_map( point ) and
			self.m_map[point[1]][point[0]] and
				self.m_map[point[1]][point[0]].id == 65)

	def is_wall( self, point ):
		if self.is_unbreachable( point[0], point[1] ):
			return True 
		
		if ( self.is_filler( point ) == True):
			return False
            
		if ( not self.is_unvisited_point( point ) ):
			return True
		item = self.m_map[point[1]][point[0]]
		if item and item.id >= 50:
			return True

		return False

	def is_obstructed( self, point ):
		row, col = point
		return (not self.is_within_map( row, col ) or self.m_map[col][row] is None or
                        self.m_map[col][row].id >= 50)

	def smooth_wall( self, row, col ):
                if not self.is_obstructed( (row, col) ):
			return

		# look at the 4 spaces around the point, up down left right                       
		# horizontal wall: if there is space above and below, but wall on either side   ---  
		#                                                                                  
		#										                                           |
		# vertical wall: if there is space to the left and right, but wall above and below |
		#                                                                                  |
	    #                                                                                    -+
		# upper right corner: if there is wall below and space above and space to the east    |
		#
        #                                                                                     +-
		# upper left corner: if there is wall below and space above and a space to the west   |
		# lower right corner: wall above, space below, wall to left, space to the right         |
		#                                                                                      -+
		# lower left corner:  wall above, space below, wall to right, space to the left     |
		#                                                                                   +-
		# up end: space on all sides except below
		# down end: space on all sides except above

		# edge cases

		pointAbove = row - 1, col
		pointBelow = row + 1, col
		pointLeft = row, col - 1
		pointRight = row, col + 1
		pointUpCatty = row-1, col-1
		pointUpKitty = row-1, col+1
		pointDownCatty = row+1, col-1
		pointDownKitty = row+1, col+1

		#print row, col
		#self.print_map()

		# handle inner corners
		# upper left corner
		if not self.is_obstructed(pointDownKitty) \
			and self.is_obstructed(pointRight) == True and self.is_obstructed(pointBelow) == True:
			self.mark2( row, col, things.Tile(55) )
			return
		# upper right corner
		elif not self.is_obstructed(pointDownCatty) \
			and self.is_obstructed(pointLeft) == True and self.is_obstructed(pointBelow) == True:
			self.mark2( row, col, things.Tile(56) )
			return
		# lower left corner
		elif not self.is_obstructed(pointUpKitty) \
			and self.is_obstructed(pointRight) == True and self.is_obstructed(pointAbove) == True:
			self.mark2( row, col, things.Tile(59) )
			return
		elif not self.is_obstructed(pointUpCatty) \
			and self.is_obstructed(pointLeft) == True and self.is_obstructed(pointAbove) == True:
			self.mark2( row, col, things.Tile(60) )
			return

		#print "points in question are: ", row, " ", col, "  above: ", pointAbove, " below: ", pointBelow, " left: ", pointLeft, " right: ", pointRight
		surrounds = (self.is_obstructed(pointAbove), self.is_obstructed(pointBelow), self.is_obstructed(pointLeft),
                             self.is_obstructed(pointRight))

		self.mark2( row, col, things.Tile(self.wall_type(surrounds)) )

	def get_id( self, row, col ):
		tile = self.m_map[col][row]
		if ( tile == None ):
			return -1
		return tile.id

	# @return a Tile id for this type of wall
	# @param surrounds a 4-tuple of booleans:
	#                  whether (above, below, left, right) walls exist
	def wall_type(self, surrounds):
		above, below, left, right = surrounds
		return 50 + ((above << 3) | (below << 2) | (left << 1) | right)

	def draw_end_cap( self, yourrow, yourcol, direction, width ):
		half = width / 2
		col = yourcol
		
		if ( direction == "north" ):
			row = yourrow + 1
			col = yourcol - width
			for x in range( width + 4 ):
				if ( self.is_unvisited( row, col+x ) ):
					self.mark2( row, col+x, things.Tile(57) )
				if ( self.is_unvisited( row+1, col+x ) ):
					self.mark2( row+1, col+x, things.Tile(64) )
		if ( direction == "south" ):
			row = yourrow - 1
			col = yourcol - width
			for x in range( width + 4 ):
				if ( self.is_unvisited( row, col+x ) ):
					self.mark2( row, col+x, things.Tile(61) )
				if ( self.is_unvisited( row-1, col+x ) ):
					self.mark2( row-1, col+x, things.Tile(64) )
		if ( direction == "east" ):
			col = yourcol - 1 
			row = yourrow - width
			for x in range( width + 4 ):
				if ( self.is_unvisited( row+x, col ) ):
					self.mark2( row+x, col, things.Tile(62) )
				if ( self.is_unvisited( row+x, col-1 ) ):
					self.mark2( row+x, col-1, things.Tile(64) )
		if ( direction == "west" ):
			col = yourcol + 1 
			row = yourrow - width
			for x in range( width + 4 ):
				if ( self.is_unvisited( row+x, col ) ):
					self.mark2( row+x, col, things.Tile(63) )
				if ( self.is_unvisited( row+x, col+1 ) ):
					self.mark2( row+x, col+1, things.Tile(64) )


	def draw_border_wall(self):
		# go around the edge and put a special wall there
		for x in range( self.m_nrows ):
			for y in range( self.m_ncols ):	
				if ( x == 0 or x == 1 or y == 0 or y == 1 or x == ( self.m_nrows - 2 ) or ( x == self.m_nrows - 1 ) \
					or y == self.m_ncols - 2 or y == self.m_ncols - 1 ):
					self.mark2( x, y, things.Tile(65) )		


	def make_maze(self):
		# before we get started, make an unbreachable wall around the outside of the map
		self.draw_border_wall()

		direction = "none"

		corridor_width = 3

		# pick a start point to begin drawing a maze
		point = self.pick_a_random_unvisited_point()
		row = point[0]
		col = point[1]
		self.mark2( row, col, things.Tile(1) )
		
		nCounter = 0
		while ( not self.map_completed() ):

#			self.print_map()

			# look ahead to the next direction
			OkaytoProceed, row2, col2, direction = self.pick_a_direction( row, col, direction, corridor_width )
			if ( nCounter == 0 ):	 # only first time
				self.mark_walls( row, col, direction, corridor_width )
				self.draw_end_cap( row, col, direction, corridor_width )
			else:
				OkaytoProceed, row2, col2, direction = self.blast_through_a_wall( row, col, direction, corridor_width )
			nCounter = nCounter + 1

			# don't mark walls until we know which direction we're going
			#self.mark_walls( row, col, direction, corridor_width )

			# go somewhere for a while
			while ( OkaytoProceed == True ):
				if ( not self.is_free_square( row2, col2 ) ):
					break	
				self.mark2( row2, col2, things.Tile(1) )
				oldrow = row2
				oldcol = col2
				olddir = direction
				OkaytoProceed, row2, col2, direction = self.pick_a_direction( row2, col2, direction, corridor_width ) 

				# mark walls after changing directions
				if ( olddir != "none" and olddir != direction ):
					row2, col2 = self.draw_corner( corridor_width, oldrow, oldcol, row2, col2, olddir, direction )
#					self.print_map()
				else:
					self.mark_walls( oldrow, oldcol, direction, corridor_width )

#				self.print_map()

			point = self.pick_a_random_path_point()
			row = point[0]
			col = point[1]
			if ( row == -1 or col == -1 ):
				break
			self.mark2( row, col, things.Tile(1) )

		self.add_rooms( corridor_width )

		self.smooth()

	def add_rooms( self, width ):
		rooms = []
		for x in range( self.m_nrooms ):
			rooms.append( self.add_room( width, rooms ) )

	def get_widest_line( self, orient ):
		longs = self.get_widest_lines( orient )
		return longs[0]

	def get_widest_lines( self, orient ):
		longest = 0
		longones = []
		if ( orient == "horizontal" ):
			for line in range( self.m_nrows) :
				# scan this line, find the widest open space
				hlen, hindex = self.find_widest_open_space( line, self.m_nrows, "horizontal")
				if ( hlen > longest ):
					longest = hlen
					longones.append( LinePair( line, hlen ) )
		 
		if ( orient == "vertical" ):
			for line in range( self.m_ncols ) :
				# scan this line, find the widest open space
				vlen, vindex = self.find_widest_open_space( line, self.m_ncols, "vertical")
				if ( vlen > longest ):
					longline = line
					longest = vlen
					longones.append( LinePair( line, vlen ) )

		longones.sort()
		longones.reverse()

		return  longones
		 
	def pick_random_path_point( self ):
		things = self.get_path_points()
		point = random.choice( things )
		return point
		
	def pick_new_random_path_point( self, rooms ):
		things = self.get_other_path_points( rooms )
		point = random.choice( things )
		return point
		
	def add_room( self, cor_width, rooms ):
		# just plunk a room down anywhere, if it covers cooridors, that's good

		width = cor_width + 4
		high = random.randrange( width, self.m_nmax_room_width )

		wide = random.randrange( width, self.m_nmax_room_width )

		point = self.pick_new_random_path_point( rooms )

		#if point is so close to edge of map that drawing the room is pointless
		# then adjust the point backwards
		if ( point[0] + wide > self.m_ncols ):
			point[0] = self.m_ncols - wide - width
		if ( point[1] + high > self.m_nrows ):
			point[1] = self.m_nrows - high  - width

		# plunk a room down, test first, must pass at least one path space
		good = False
		for x in range( wide ):
			if ( good == True ):
				break
			for y in range( high ):
				if ( self.is_path( point[0] + x, point[1] + y ) ):
					good = True
					break
		if (good):
			self.draw_room( wide, high, point[0], point[1] )	

		return [ point[0], point[1], point[0] + wide, point[1] + high ]
				

	def draw_room( self, hlen, vlen, hindex, vindex ):
		count = 0
		for xindex in range( hlen ):
			x = xindex + hindex
			count = count + 1
			for yindex in range( vlen ):
				y = yindex + vindex
				if self.is_path( x, y ) or self.is_unbreachable( x, y ): # leave existing paths there
					continue
				if ( x == 0 or x == 1 or x == hlen - 2 or x == hlen - 1 or y == 0 or y == 1 or y == vlen - 2 or y == vlen - 1 ):
					self.mark2( x, y, things.Tile( 62 ) ) # just pick some wall tile, smooth will will straighten it out
				else:
					self.mark2( x, y, things.Tile(1) )
		return True
				
	def find_widest_open_space( self, linenum, length, orient ):
		longest = 0
		longest_index = -1
		len = 0;
		index = 0;
		if ( orient == "horizontal" ):
			for col in range( length ): 
				if self.is_unvisited( linenum, col ):
					if ( index == -1 ):
						index = col 
					#	print "new unvisited index is", col 
					len = len + 1
					if ( len > longest ):
						longest = len
						longest_index = index
					#	print "new longest lenght is",  len
					#	print "new longest index is", longest_index
				else:
					len = 0
					index = -1
						
		if ( orient == "vertical" ):
			for row in range( length ): 
				if self.is_unvisited( row, linenum ):
					if ( index == -1 ):
						index = row
					len = len + 1
					if ( len > longest ):
						longest = len
						longest_index = index
				else:
					len = 0
					index = -1
						
		return longest, longest_index
							
						
				

	def smooth( self ):
		for row in range(self.m_nrows):
			for col in range(self.m_ncols):
				self.smooth_wall( row, col )

	
	def print_map(self):
		# make a text representation of the map
		result = "  "
		for col in range(self.m_ncols):
			result = result + str(col%10)
		print result

		for row in range(self.m_nrows):
			result = ""
			if ( row < 10 ):
				result = result + " "
			result = result + str(row)
			for col in range(self.m_ncols):
				square = self.m_map[col][row]
				if square is None:
					result = result + '0'
				else:
					result = result + str(square)
			print result
		result = "  "
		for col in range(self.m_ncols):
			result = result + str(col%10)
		print result


	def __str__(self):
		# make a text representation of the map
		result = []
		for row in range(self.m_nrows):
			row_result = []

		for col in range(self.m_ncols):
			square = self.m_map[col][row]

			if square is None:
				row_result.append('0')
			else:
				row_result.append(str(square))
			result.append(''.join(row_result))
		return '\n'.join(result)

		
#map = Map( 64, 64, 0.4, 25, 12, 5, 8 )
#map = Map( 44, 44, 0.4, 15, 12, 5, 12 )
#map.make_maze( )
#pdb.runcall( map.make_maze )
#map.print_map()

	
