### Main server game module

import pygame

import mapgenerator
import things
import player
import monster

import random
import threading
import urllib

global_lock = threading.Lock()

have_dm = False

player_crystals = 0
dm_crystals = 0

cur_id = 1L
id_lock = threading.Lock()

def get_id():
    global cur_id
    
    id_lock.acquire()
    try:
        result = cur_id
        cur_id += 1
    finally:
        id_lock.release()
    return result

chatlines = [(0,""),(0,""),(0,"Welcome to Pencil Quest")]

def setChat(pid,chatmsg):
    if chatmsg != "":
        chatlines[0] = chatlines[1]
        chatlines[1] = chatlines[2]
        chatlines[2] = (pid,urllib.unquote(chatmsg))

    if chatmsg == "l33t":
        chatlines[2] = (pid,"I am not very l337 since I spelled it wrong.")

class Map:
    debug = 0
    
    def __init__(self, mapsize):
	#                                         sparseness  turn freq  max rooms  min rooms  max room width 
        maze = mapgenerator.Map(mapsize, mapsize, 0.30,       10,        10,        10,        8 )
        maze.make_maze()
        if self.debug:
            print repr(maze.m_map)
            print maze
        self.grid = maze.m_map
        
        self.things = {}
        self.players = []

        self.positions = {}
        self.make_positions()

        self.generator = None
        self.last_generate = 0

        # Rect in tile coordinates
        self.rect = pygame.Rect((0, 0),
                                (len(self.grid) - 1, len(self.grid[0]) - 1))

        self.who_won = ''

    def make_positions(self):
        for key, value in self.things.items():
            self.positions[value.position] = value

    def get_dm(self):
        # XXX Probably some serious multi-threading issues here if
        # multiple people try to be the DM at once.
        global have_dm
        if have_dm: 
            return False
        else:
            have_dm = True
            return True
        
    def get_pc(self,character):
        # Generate new player and assign
        while 1:
            position = (random.randint(0, self.rect.width),
                        random.randint(0, self.rect.height))
            if self.movable_space(position[0], position[1]):
                break
        if character == 'knight':
            newPlayer = player.Knight(get_id(), position, character)
        elif character == 'mage':
            newPlayer = player.Mage(get_id(), position, character)
        elif character == 'elf':
            newPlayer = player.Elf(get_id(), position, character)
        self.players.append(newPlayer)
        self.things[newPlayer.id] = newPlayer
        self.positions[newPlayer.position] = newPlayer
        return newPlayer.id

    def get_grid(self):
        return map.grid

    def get_state(self):
        return (map.things, chatlines, map.who_won)
        
    def movable_space(self, x, y):
        if x < 0 or y < 0: return False # -1 is valid index so no IndexError
        try: space = self.grid[x][y]
        except IndexError: return False
        return ((space.id < 50) and (self.positions.get((x,y)) is None))

    def hopefully_movable_space(self, x, y):
        if x < 0 or y < 0: return False # -1 is valid index so no IndexError
        try: space = self.grid[x][y]
        except IndexError: return False
        if space.id >= 50: return False
        thing = self.positions.get((x, y))
        if thing is None: return True
        return isinstance(thing, things.Creature)

    def command_monster(self, pid, x, y, w, h):
        thing = self.things[pid]
        if isinstance(thing, things.Monster):
            thing.set_wander_rectangle(pygame.Rect(x,y,w,h))

    def move(self, thing, delta_x, delta_y):
        global player_crystals, dm_crystals
        
        if type(thing) in (type(0), type(0L)):
            thing = self.things[thing]

        if isinstance(thing, things.Creature) and thing.engaged():
            return False
            
        x, y = thing.position
        
        x += delta_x
        x = max(0, x)
        x = min(x, mapsize - 1)
        
        y += delta_y
        y = max(0, y)
        y = min(y, mapsize - 1)

        # later, dungeon master can come through and re-own this somehow
        if isinstance(thing, things.Player):
            thing2 = self.positions.get((x, y))
            if isinstance(thing2, things.Crystal):
                thing.own(thing2)
                thing2.owned_by( thing )
                if thing2.owner_team != 'players':
                    if thing2.owner_team == 'dm':
                        dm_crystals -= 1
                    thing2.owner_team = 'players'
                    player_crystals += 1
                    if player_crystals >= 4:
                        map.who_won = 'player'
                return True
            elif isinstance(thing2, things.Item):
                thing.pick_item(thing2)
        
        # don't walk into things, and walls which have tile.id >= 50
        if (self.movable_space(x,y)):
            try: del self.positions[thing.position]
            except KeyError: pass
            thing.position = x, y
            self.positions[thing.position] = thing

            if thing.debug:
                print "%s) --> %s" % (thing.id, thing.position)

            if isinstance(thing, things.Creature):
                thing.check_engagement()

            return True
        elif isinstance(thing, things.Player) and delta_x and delta_y:
            # Try a flat move although they are asking for diagnonal
            # to stop the player from getting stuck on walls.
            for new_x, new_y in ((delta_x,0),(0,delta_y)):
                if self.move(thing,new_x,new_y):
                    return True
        return False

    def item_action(self, id, action):
        if action == "pick":
            self.things[id].pick_item()

    def sprinkle_monsters(self):
        choices = []
        for y in range(mapsize):
            for x in range(mapsize):
                if self.movable_space(x,y) and \
                   not self.positions.has_key((x,y)) and \
                   random.randint(0,99) < 3:
                    choices.append((x,y))
        id = get_id()
        pos = random.choice(choices)
        monster = things.Dragon(id,pos)
        self.things[id] = monster
        self.positions[pos] = monster
        choices.remove(pos)
        id = get_id()
        pos = random.choice(choices)
        monster = things.CrystalMaster(id,pos)
        self.things[id] = monster
        self.positions[pos] = monster
        choices.remove(pos)
        for i in range(2):
            id = get_id()
            pos = random.choice(choices)
            monster = things.Demon(id,pos)
            self.things[id] = monster
            self.positions[pos] = monster
            choices.remove(pos)
        for i in range(3):
            id = get_id()
            pos = random.choice(choices)
            monster = things.Ogre(id,pos)
            self.things[id] = monster
            self.positions[pos] = monster
            choices.remove(pos)
        for i in range(4):
            id = get_id()
            pos = random.choice(choices)
            monster = things.Goblin(id,pos)
            self.things[id] = monster
            self.positions[pos] = monster
            choices.remove(pos)
        for i in range(5):
            id = get_id()
            pos = random.choice(choices)
            monster = things.Wolf(id,pos)
            self.things[id] = monster
            self.positions[pos] = monster
            choices.remove(pos)
        for i in range(6):
            id = get_id()
            pos = random.choice(choices)
            monster = things.Python(id,pos)
            self.things[id] = monster
            self.positions[pos] = monster
            choices.remove(pos)

    def create_generator(self):
        choices = []
        for y in range(mapsize):
            for x in range(mapsize):
                if self.movable_space(x,y) and \
                   not self.positions.has_key((x,y)) and \
                   random.randint(0,99) < 3:
                    choices.append((x,y))
        id = get_id()
        pos = random.choice(choices)
        generator = things.Thing(id,pos,'generator')
        self.generator = id
        self.things[id] = generator
        self.positions[pos] = generator

    def maybe_generate_monsters(self):
        number = random.randint(0,10000)
        if number > self.last_generate:
            self.last_generate += 1
            return
        if not self.generator: return
        self.last_generate = 0
        gen_pos = self.things[self.generator].position
        monster_class = random.choice([
            things.Python,
            things.Wolf,
            things.Goblin,
            things.Ogre,
            things.Demon,
            things.Dragon,
            ])
        for x in range(gen_pos[0]-1,gen_pos[0]+2):
            for y in range(gen_pos[1]-1,gen_pos[1]+2):
                if (x,y) == gen_pos: continue
                if self.movable_space(x,y):
                    id = get_id()
                    monster = monster_class(id,(x,y))
                    self.things[id] = monster
                    self.positions[monster.position] = monster
                    return

    def sprinkle_items(self):
        choices = []
        for y in range(mapsize):
            for x in range(mapsize):
                if self.movable_space(x,y) and \
                   not self.positions.has_key((x,y)) and \
                   random.randint(0,99) < 3:
                    choices.append((x,y))
        for i in range(5):
            id = get_id()
            pos = random.choice(choices)
            item = things.Armor(id,pos, 'armor', 'chest',0,-2,0,5)
            self.things[id] = item
            self.positions[pos] = item
            choices.remove(pos)
        for i in range(5):
            id = get_id()
            pos = random.choice(choices)
            item = things.Weapon(id,pos, 'weapon', 'chest',5,3,0,0)
            self.things[id] = item
            self.positions[pos] = item
            choices.remove(pos)
        for i in range(5):
            id = get_id()
            pos = random.choice(choices)
            item = things.Special(id,pos, 'special', 'chest',0,0,5,1)
            self.things[id] = item
            self.positions[pos] = item
            choices.remove(pos)
        colors = ('blue', 'green', 'purple', 'red' )
        for color in colors:
            id = get_id()
            pos = random.choice(choices)
            item = things.Crystal(id, pos, 'crystals', 'chest', color, None )
            self.things[id] = item
            self.positions[pos] = item
            choices.remove(pos)

mapsize = 64
map = Map(mapsize)
map.sprinkle_monsters()
map.sprinkle_items()
map.create_generator()
